#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # NeuralTreeNet et ONNX
# 
# La conversion d'un arbre de décision au format ONNX peut créer des différences entre le modèle original et le modèle converti (voir [Issues when switching to float](http://www.xavierdupre.fr/app/onnxcustom/helpsphinx/gyexamples/plot_ebegin_float_double.html). Le problème vient d'un changement de type, les seuils de décisions sont arrondis au float32 le plus proche de leur valeur en float64 (double). Qu'advient-il si l'arbre de décision est converti en réseau de neurones d'abord.
# 
# L'approximation des seuils de décision ne change pas grand chose dans la majorité des cas. Cependant, il est possible que la comparaison d'une variable à un seuil de décision arrondi soit l'opposé de celle avec le seuil non arrondi. Dans ce cas, la décision suit un chemin différent dans l'arbre.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')


# In[3]:


get_ipython().run_line_magic('load_ext', 'mlprodict')


# ## Jeu de données
# 
# On construit un jeu de donnée aléatoire.

# In[4]:


import numpy

X = numpy.random.randn(10000, 10)
y = X.sum(axis=1) / X.shape[1]
X = X.astype(numpy.float64)
y = y.astype(numpy.float64)


# In[5]:


middle = X.shape[0] // 2
X_train, X_test = X[:middle], X[middle:]
y_train, y_test = y[:middle], y[middle:]


# ## Partie scikit-learn

# ### Caler un arbre de décision

# In[6]:


from sklearn.tree import DecisionTreeRegressor

tree = DecisionTreeRegressor(max_depth=7)
tree.fit(X_train, y_train)
tree.score(X_train, y_train), tree.score(X_test, y_test)


# In[7]:


from sklearn.metrics import r2_score
r2_score(y_test, tree.predict(X_test))


# La profondeur de l'arbre est insuffisante mais ce n'est pas ce qui nous intéresse ici.

# ### Conversion au format ONNX

# In[8]:


from mlprodict.onnx_conv import to_onnx

onx = to_onnx(tree, X[:1].astype(numpy.float32))


# In[9]:


from mlprodict.onnxrt import OnnxInference

x_exp = X_test

oinf = OnnxInference(onx, runtime='onnxruntime1')
expected = tree.predict(x_exp)

got = oinf.run({'X': x_exp.astype(numpy.float32)})['variable']
numpy.abs(got - expected).max()


# In[10]:


from mlprodict.plotting.text_plot import onnx_simple_text_plot
print(onnx_simple_text_plot(onx))


# ## Après la conversion en un réseau de neurones

# ### Conversion en un réseau de neurones
# 
# Un paramètre permet de faire varier la pente des fonctions sigmoïdes utilisées.

# In[11]:


from tqdm import tqdm
from pandas import DataFrame
from mlstatpy.ml.neural_tree import NeuralTreeNet

xe = x_exp[:500]
expected = tree.predict(xe)

data = []
trees = {}
for i in tqdm([0.3, 0.4, 0.5, 0.7, 0.9, 1] + list(range(5, 61, 5))):
    root = NeuralTreeNet.create_from_tree(tree, k=i, arch='compact')
    got = root.predict(xe)[:, -1]
    me = numpy.abs(got - expected).mean()
    mx = numpy.abs(got - expected).max()
    obs = dict(k=i, max=mx, mean=me)
    data.append(obs)
    trees[i] = root


# In[12]:


df = DataFrame(data)
df


# In[13]:


df.set_index('k').plot(title="Précision de la conversion\nen réseau de neurones");


# L'erreur est meilleure mais il faudrait recommencer l'expérience plusieurs fois avant de pouvoir conclure afin d'obtenir un interval de confiance pour le même type de jeu de données. Ce sera pour une autre fois. Le résultat dépend du jeu de données et surtout de la proximité des seuils de décisions. Néanmoins, on calcule l'erreur sur l'ensemble de la base de test. Celle-ci a été tronquée pour aller plus vite.

# In[14]:


expected = tree.predict(x_exp)
got = trees[50].predict(x_exp)[:, -1]
numpy.abs(got - expected).max(), numpy.abs(got - expected).mean()


# On voit que l'erreur peut-être très grande. Elle reste néanmoins plus petite que l'erreur de conversion introduite par ONNX.

# ### Conversion au format ONNX
# 
# On crée tout d'abord une classe qui suit l'API de scikit-learn et qui englobe l'arbre qui vient d'être créé qui sera ensuite convertit en ONNX.

# In[15]:


from mlstatpy.ml.neural_tree import NeuralTreeNetRegressor

reg = NeuralTreeNetRegressor(trees[50])
onx2 = to_onnx(reg, X[:1].astype(numpy.float32))


# In[16]:


print(onnx_simple_text_plot(onx2))


# In[17]:


oinf2 = OnnxInference(onx2, runtime='onnxruntime1')
expected = tree.predict(x_exp)

got = oinf2.run({'X': x_exp.astype(numpy.float32)})['variable']
numpy.abs(got - expected).max()


# L'erreur est la même.

# ## Temps de calcul

# In[18]:


x_exp32 = x_exp.astype(numpy.float32)


# Tout d'abord le temps de calcul pour scikit-learn.

# In[19]:


get_ipython().run_line_magic('timeit', 'tree.predict(x_exp32)')


# Le temps de calcul pour l'arbre de décision au format ONNX.

# In[20]:


get_ipython().run_line_magic('timeit', "oinf.run({'X': x_exp32})['variable']")


# Et le temps de calcul pour le réseau de neurones au format ONNX.m

# In[21]:


get_ipython().run_line_magic('timeit', "oinf2.run({'X': x_exp32})['variable']")


# Ce temps de calcul très long est attendu car le modèle contient une multiplication de matrice très grande et surtout que tous les seuils de l'arbre sont calculés pour chaque observation. Là où l'implémentation de l'arbre de décision calcule *d* seuils, la profondeur de l'arbre, la nouvelle implémentation calcule tous les seuils soit $2^d$ pour chaque feuille. Il y a $2^d$ feuilles. Même en étant sparse, on peut réduire les calculs à $d * 2^d$ ce qui fait encore beaucoup de calculs inutiles.

# In[22]:


for node in trees[50].nodes:
    print(node.coef.shape, node.bias.shape)


# Cela dit, la plus grande matrice est creuse, elle peut être réduite considérablement.

# In[23]:


from scipy.sparse import csr_matrix

for node in trees[50].nodes:
    csr = csr_matrix(node.coef)
    print(f"coef.shape={node.coef.shape}, size dense={node.coef.size}, "
          f"size sparse={csr.size}, ratio={csr.size / node.coef.size}")


# In[24]:


r = numpy.random.randn(trees[50].nodes[1].coef.shape[0])
mat = trees[50].nodes[1].coef
get_ipython().run_line_magic('timeit', 'mat @ r')


# In[25]:


csr = csr_matrix(mat)
get_ipython().run_line_magic('timeit', 'csr @ r')


# Ce serait beaucoup plus rapide avec une matrice sparse et d'autant plus rapide que l'arbre est profond. Le modèle ONNX se décompose comme suit.

# In[26]:


print(onnx_simple_text_plot(onx2))


# Voyons comment le temps de calcul se répartit.

# In[27]:


oinfpr = OnnxInference(onx2, runtime="onnxruntime1",
                      runtime_options={"enable_profiling": True})
for i in range(0, 43):
    oinfpr.run({"X": x_exp32})


# In[28]:


df = oinfpr.get_profiling(as_df=True)
df


# In[29]:


set(df['args_provider'])


# In[30]:


dfp = df[df.args_provider == 'CPUExecutionProvider'].copy()
dfp['name'] = dfp['name'].apply(lambda s: s.replace("_kernel_time", ""))
gr_dur = dfp[['dur', "args_op_name", "name"]].groupby(["args_op_name", "name"]).sum().sort_values('dur')
gr_dur


# In[31]:


gr_n = dfp[['dur', "args_op_name", "name"]].groupby(["args_op_name", "name"]).count().sort_values('dur')
gr_n = gr_n.loc[gr_dur.index, :]
gr_n


# In[32]:


import matplotlib.pyplot as plt

fig, ax = plt.subplots(1, 2, figsize=(12, 4))
gr_dur.plot.barh(ax=ax[0])
gr_n.plot.barh(ax=ax[1])
ax[0].set_title("duration")
ax[1].set_title("n occurences");


# onnxruntime passe principalement son temps dans un produit matriciel. On vérifie plus précisément.

# In[33]:


df[(df.args_op_name == 'Gemm') & (df.dur > 0)].sort_values('dur', ascending=False).head(n=2).T


# C'est un produit matriciel d'environ *5000x800* par *800x800*.

# In[34]:


gr_dur / gr_dur.dur.sum()


# In[35]:


r = (gr_dur / gr_dur.dur.sum()).dur.max()
r


# Il occupe 82% du temps. et d'après l'expérience précédente, son temps d'éxecution peut-être réduit par 10 en le remplaçant par une matrice sparse. Cela ne suffira pas pour accélerer le temps de calcul de ce réseau de neurones. Il est 84 ms comparé à 247 µs pour l'arbre de décision. Avec cette optimisation, il pourrait passer de :

# In[36]:


t = 3.75  # ms
t * (1 - r) + r * t / 12


# Soit une réduction du temps de calcul. Ce n'est pas mal mais pas assez.

# ## Hummingbird
# 
# [hummingbird](https://github.com/microsoft/hummingbird) est une librairie qui convertit un arbre de décision en réseau de neurones. Voyons ses performances.

# In[37]:


from hummingbird.ml import convert

model = convert(tree, 'torch')

expected = tree.predict(x_exp)
got = model.predict(x_exp)
numpy.abs(got - expected).max(), numpy.abs(got - expected).mean()


# Le résultat est beaucoup plus fidèle au modèle.

# In[38]:


get_ipython().run_line_magic('timeit', 'model.predict(x_exp)')


# Il reste plus lent mais beaucoup plus rapide que la solution manuelle proposée dans les précédents paragraphes. Il contient un attribut `model`.

# In[39]:


from torch.nn import Module
isinstance(model.model, Module)


# On convertit ce modèle au format ONNX.

# In[40]:


import torch.onnx

x = torch.randn(x_exp.shape[0], x_exp.shape[1], requires_grad=True)
torch.onnx.export(model.model, x, 'tree_torch.onnx', opset_version=15,                         
                  input_names=['X'], output_names=['variable'],
                  dynamic_axes={
                      'X' : {0 : 'batch_size'},
                      'variable' : {0 : 'batch_size'}})


# In[41]:


import onnx

onxh = onnx.load('tree_torch.onnx')


# In[42]:


print(onnx_simple_text_plot(onxh, raise_exc=False))


# In[43]:


get_ipython().run_line_magic('onnxview', 'onxh')


# La librairie réimplémente la décision d'un arbre décision à partir d'un produit matriciel pour chaque niveau de l'arbre. Tous les seuils sont évalués. Les matrices n'ont pas besoin d'être sparses car les features nécessaires sont récupérées. Le seuil de décision est implémenté avec un test et non une sigmoïde. Ce modèle est donc identique en terme de prédiction au modèle initial.

# In[44]:


oinfh = OnnxInference(onxh, runtime='onnxruntime1')
expected = tree.predict(x_exp)

got = oinfh.run({'X': x_exp.astype(numpy.float32)})['variable']
numpy.abs(got - expected).max()


# La conversion reste imparfaite également.

# In[45]:


get_ipython().run_line_magic('timeit', "oinfh.run({'X': x_exp32})['variable']")


# Et le temps de calcul est aussi plus long.

# ## Apprentissage
# 
# L'idée derrière tout cela est aussi de pouvoir réestimer les coefficients du réseau de neurones une fois converti.

# In[46]:


x_train = X_train[:100]
expected = tree.predict(x_train)
reg = NeuralTreeNetRegressor(trees[1], verbose=1, max_iter=10, lr=1e-4)


# In[47]:


got = reg.predict(x_train)
numpy.abs(got - expected).max(), numpy.abs(got - expected).mean()


# La différence est grande.

# In[48]:


reg.fit(x_train, expected)


# In[49]:


got = reg.predict(x_train)
numpy.abs(got - expected).max(), numpy.abs(got - expected).mean()


# Ca ne marche pas aussi bien que prévu. Il faudrait sans doute plusieurs itérations et jouer avec les paramètres d'apprentissage.

# In[50]: